%
% Set of working functions for "tsd" - Tanabe-Sugano diagrams generator
%
% Author: Adam Drzewiecki
% e-mail: A.Drzewiecki@proton.if.uz.zgora.pl
% WWW: http://www.if.uz.zgora.pl/~adam
%
% Last modification: 07 Dec 2012
%

% Dear Octave, it is not a function file! :)
0;

%
% Helper functions section
%


% Checker for n - number of electrons
function check_n(n)
    if (nargin ~= 1)
	error('check_n(n)');
    end

    if (!isscalar(n) || !isreal(n))
	error('n must be integer');
    elseif (n < 1 || n > 9)
	error('n must be 1-9 integer number');
    end
end

% Checker for f - file identifier
function check_f(f)
    if (nargin ~= 1)
	error('check_f(f)');
    end

    if (!isscalar(f) || !isreal(f))
	error('file identifier (f) must be integer');
    elseif (f == -1)
	error('file identifier must be valid');
    end
end

% Checker for config - structure describing set of crystal field terms
function check_config(config)
    if (nargin ~= 1)
	error('check_config(config)');
    end

    if (!isstruct(config))
	error('config must be a structure');
    end
end

% Checker for ctb - C/B ratio
function check_ctb(ctb)
    if (nargin ~= 1)
	error('check_ctb(ctb)');
    end

    if (!isscalar(ctb) || !isreal(ctb))
	error('C/B ratio must be a real scalar');
    elseif (ctb < 0)
	error('C/B ratio must be a non-negative number');
    end
end

% Checker for separator - string using for separation of columns in results table
function check_separator(separator)
    if (nargin ~= 1)
	error('check_separator(separator)');
    end

    if (!ischar(separator))
	error('separator must be a string');
    end
end

% Checker for x coordinate
function check_x(x)
    if (nargin ~= 1)
	error('check_x(x)');
    end

    if (!isscalar(x) || !isreal(x))
	error('x must be a real scalar value');
    end
end

% Checker for number of points
function check_points(points)
    if (nargin ~= 1)
	error('check_points(points)');
    end

    if (!isscalar(points) || !isreal(points))
	error('Number of points must be a real scalar');
    elseif (points < 1)
	error('Number of points must be greater or equal 1');
    end
end

%
% Working functions section
%

%
% Function loads information about level splitting of d(n) configuration
%
function config = load_configuration(n)
    % Check call argument
    if (nargin ~= 1)
	error('load_configuration(n)');
    end

    check_n(n);

    % Load diagram states configuration
    switch n
	case 1
	    config = d1_levels();

	case 2
	    config = d2_levels();

	case 3
	    config = d3_levels();

	case 4
	    config = d4_levels();

	case 5
	    config = d5_levels();

	case 6
	    config = d6_levels();

	case 7
	    config = d7_levels();

	case 8
	    config = d8_levels();

	case 9
	    config = d9_levels();

    otherwise
	error('Wrong diagram number');

    end
end

%
% Function generates string form of crystal field term description
%
function td = term_description(m, s, t)
    if (nargin ~= 3)
	error('term_description(m, s, t)');
    end

    if (!isscalar(m) || !isreal(m))
	error('multiplicity (m) must be integer number');
    end

    if (!ischar(s))
	error('symmetry (s) must be a string');
    end

    if (!ischar(t))
	error('atomic term (t) must be a string');
    end

    % Prepare output string
    td = sprintf('%i%s(%s)', m, s, t);
end

%
% Function writes column of number n description text t to the file f
%
function column_description(f, col, txt)
    % Check call arguments
    if (nargin ~= 3)
	error('column_description(f, n, t)');
    end

    check_f(f);

    if (!isscalar(col) || !isreal(col))
	error('column number (col) must be integer');
    end

    if (!ischar(txt))
	error('column text (txt) description must be a string');
    end

    % Write formatted line with column number and text description to file
    fprintf(f, '# %3i) %s\n', col, txt);
end

%
% Function writes descriptive info about crystal field terms od d(n) electron configuration
%
function write_descriptions(f, config)
    % Check call arguments
    if (nargin ~= 2)
	error('write_descriptions(f, config)');
    end

    check_f(f);
    check_config(config);

    % Description of the first colum
    column_description(f, 1, 'Delta/B');

    % Starting from column index equal two:
    col = 2;
    % Index i-th configuration matrix of crystal field term
    for i = 1:columns(config)
	%... get description of it
	s = config(i);

	% For every atomic term included in current symmetry term
	for t = 1:columns(s.terms)
	    % Write description to file
	    column_description(f, col, term_description(s.multiplicity, s.symmetry, s.terms{t}));
	    % Prepare next column number
	    col++;
	end
    end
end

%
% Function writes header section to output file
%
function write_header(f, n, config, ctb)
    % Check call arguments
    if (nargin < 3 || nargin > 4)
	error('write_header(f, n, config, ctb)');
    end

    check_f(f);
    check_n(n);
    check_config(config);

    if(nargin == 4)
	check_ctb(ctb);
    end

    % Write some spacing
    fprintf(f, '#\n');

    % Write header title line
    if (n >= 2 && n <= 8)
	fprintf(f, '# d%i configuration Tanabe-Sugano diagram (C/B = %f) generated by \"tsd\" program\n', n, ctb);
    elseif (n == 1 || n == 9)
	fprintf(f, '# d%i configuration Tanabe-Sugano diagram generated by \"tsd\" program\n', n);
    end

    % Write some spacing and column description intro
    fprintf(f, '#\n');
    fprintf(f, '# Values:\n');
    fprintf(f, '#\n');

    % Prepare column descriptions according to configuration info
    write_descriptions(f, config);

    % Write ending of header section
    fprintf(f, '#\n');
end

%
% Function writes term labels as a results table header
%
function write_labels(f, config, separator)
    % Check call arguments
    if (nargin < 2 || nargin > 3)
	error('write_labels(f, config, separator)');
    end

    check_f(f);
    check_config(config);

    % If not given, set up default separator value
    if (nargin ~= 3)
	seperator = ' ';
    end

    check_separator(separator);

    % Write x coordinate description as a first position of a list
    fprintf(f, 'Delta/B%s', separator);

    % Get number of groups
    groups = columns(config);

    % For every group of terms with the same symmetry:
    for i = 1:groups
	% Check if it is the last group of terms.
	last_group = i == groups;

	% Get information about current (i-th) group
	s = config(i);

	% Get the number of atomic terms
	terms = columns(s.terms);

	% For every atomic term in the list:
	for j = 1:terms
	    % Check if it is the last term in the group.
	    last_term = j == terms;

	    % If it is not global last position on the list, write it with separator
	    if (!last_group || !last_term)
		fprintf(f, '%s%s', term_description(s.multiplicity, s.symmetry, s.terms{j}), separator);
	    % Last element of the list is written without separator
	    else
		fprintf(f, '%s', term_description(s.multiplicity, s.symmetry, s.terms{j}));
	    end
	end
    end

    % Write newline character at the end
    fprintf(f, '\n');
end

%
% Function calculates eigenvalues of energy matrix
%
function [mca, E0] = energies(config, x, ctb)
    % Check call arguments
    if(nargin < 2 || nargin > 3)
	error('energies(ms, x, ctb)');
    end

    check_config(config);
    check_x(x);

    if (nargin == 3)
	check_ctb(ctb);
    end

    % Get the number of groups of CF terms
    groups = columns(config);

    % Prepare empty cell array: one row, matrix number columns
    mca = cell(1, groups);

    % Prepare variable for searching of the lowest energy level
    E0 = 0.0;

    % For every matrix in a set:
    for i = 1:groups;
	% Prepare matrix:
	if (nargin > 2)
	    % Compute current term matrix values for the given x and CTB
	    m = config(i).matrix(x, ctb);
	else
	    % Compute current term matrix values for the given x (without CTB, 1 or 9 case)
	    m = config(i).matrix(x);
	end
	
	% Find eigenvalues vector of calculated matrix and store it in mca cell array
	mca{i} = eig(m);

	% Calculate lowest energy of current set of matrix eigenvalues
	Emin = min(mca{i});

	% Global selection:
	% If it is 2nd or further term matrix in a sequence
	if(i > 1)
	    % If we have a lowest value than known minimum, store it as a new lowest value
	    if(Emin < E0)
		E0 = Emin;
	    end
	% Minimum of the first matrix is just stored as a lowest known value
	else
	    E0 = Emin;
	end
    end
end

%
% Function subtracts shift from all values in cell array
%
function rca = remove_base_level(ca, shift)
    % Check call arguments
    if (nargin ~= 2)
	error('remove_base_level(ca, shift)');
    end

    if (!iscell(ca))
	error('ca must be a cell array');
    end

    if (!isscalar(shift) || !isreal(shift))
	error('shift must be a real scalar value');
    end

    % Get the number of groups of CF terms
    groups = columns(ca);

    % Prepare empty return values cell array
    rca = cell(size(ca));

    % For every set of matrix eigenvalues
    for i = 1:groups
	% Subtract shift vector form original eigenvalues vector and store result
	rca{i} = ca{i} - shift*ones(size(ca{i}));
    end
end

%
% Function writes all of the energy eigenvalues sepasrated by the given separator
%
function write_levels(f, ca, separator)
    % Check call arguments
    if (nargin < 2 || nargin > 3)
	error('write_levels(f, ca, separator)');
    end

    check_f(f);

    if (!iscell(ca))
	error('ca must be a cell array');
    end

    % If not given, set up default separator value
    if (nargin ~= 3)
	separator = ' ';
    endif

    check_separator(separator);

    % Get the number of groups
    groups = columns(ca);

    % For each group of eigenvalues:
    for i = 1:groups
	% Is this the last group of terms?
	last_group = i == groups;
	
	% Get the number of terms
	terms = rows(ca{i});
	
	% For each eigenvalue in the group:
	for j = 1:terms
	    % Is this the last term in current group?
	    last_term = j == terms;

	    % Write eigenvalue with the following separator if it is not last to be written
	    if (!last_group || !last_term)
		fprintf(f, '%f%s', ca{i}(j), separator);
	    % Last eigenvalue is written without sepator
	    else
		fprintf(f, '%f', ca{i}(j));
	    end
	end
    end

    % Finish row with newline character
    fprintf(f, '\n');
end

%
% Main working function
%
function write_diagram(f, separator, n, xmin, xmax, points, ctb)
    % Function must be called with 6 or 7 arguments, depending on n
    if (nargin < 6 || nargin > 7)
	error('write_diagram(f, separator, n, xmin, xmax, points [, ctb])');
    end

    check_f(f);
    check_separator(separator);
    check_n(n);
    check_x(xmin);
    check_x(xmax);
    check_points(points);

    % For d2-d8 diagrams C/B ratio is required
    if (n >= 2 && n <= 8)
	if (nargin == 7)
	    check_ctb(ctb);
	else
	    error('write_diagram() call for d2-d8 diagrams require C/B ratio');
	end
    end

    % Get electron level configuration data
    config = load_configuration(n);

    % Write header of the file with (d2-d8) or without (d1, d9) C/B ratio information
    if (n >= 2 && n <= 8)
	write_header(f, n, config, ctb);
    elseif (n == 1 || n == 9)
	write_header(f, n, config);
    end

    % Write term labels as a header row in the table of results
    write_labels(f, config, separator);

    % For every Delta/B value named as x...
    for x = linspace(xmin, xmax, points)
	% Write x coordinate
	fprintf(f, '%f%s', x, separator);

	% Calculate levels and find base level of energy
	if (n >= 2 && n <= 8)
	    [energy_cell_array, E0] = energies(config, x, ctb);
	elseif (n == 1 || n == 9)
	    [energy_cell_array, E0] = energies(config, x);
	end

	% Set base level to zero
	energy_cell_array = remove_base_level(energy_cell_array, E0);

	% Write all of energy levels to the file
	write_levels(f, energy_cell_array, separator);
    end
end
